# Main program for paper "Online Computation of Reduced Egonet Features
# for Anomaly Detection in Bank Transactions Graphs", by C-E.Zica, B.Dumitrescu

# Bogdan Dumitrescu & Cristian-Enache Zica 28.05.2023

# Copyright: You can use the programs at will, with no guarantee whatsoever.
# Please cite the paper if you use these programs.

import networkx as nx
import numpy as np
import online_egonet_features as oef
import pandas as pd
import time

# source file paths for historical graph, basic features,
# derived features and incremental transactions

historical_graph_path = 'Libra_historical_graph.csv'
historical_basic_features_path = 'Libra_historical_basic_features.csv'
historical_derived_features_path = 'Libra_historical_derived_features.csv'
incremental_transaction_path = 'Libra_incremental_transactions.csv'

# paths for updated basic features and time auditing
updated_basic_features_path = 'Libra_basic_features_updated.csv'
audit_path = 'elapsed_time.csv'

audit_columns = ["transaction_number", "id_source", "id_destination", "elapsed_time"]

# read historical graph, basic features, derived features and list of transactions
G = pd.read_csv(historical_graph_path)
Xf = pd.read_csv(historical_basic_features_path)
Xf_d = pd.read_csv(historical_derived_features_path)
trans_df = pd.read_csv(incremental_transaction_path)

# transform graph from pandas to networkx DiGraph
G = nx.from_pandas_edgelist(df=G, source='id_source', target='id_destination',
                            edge_attr=True, create_using=nx.DiGraph)

# how many transactions to process
# can go from 1 (pure online behavior)
# to max number of transactions in the file
max_number_of_transactions = len(trans_df)

for it in range(max_number_of_transactions):
    id_source = trans_df.iloc[it, :]["id_source"]
    id_destination = trans_df.iloc[it, :]["id_destination"]

    # for each transaction in the file
    # call online_update_features_and_graph with graph, basic features,
    # derived features and transactions to get the updated version of the first 3 variables
    start_time = time.time()
    G, Xf, Xf_d = oef.online_update_features_and_graph(G, trans_df.iloc[it, :], Xf, Xf_d)
    stop_time = time.time()

    elapsed_time_seconds = stop_time - start_time

    arr = np.array([[it,
                     id_source,
                     id_destination,
                     elapsed_time_seconds]])

    # store the elapsed time information
    if it == 0:
        audit_arr = arr
    else:
        audit_arr = np.append(audit_arr, arr, axis=0)

    if it % 100 == 0:
        print(f"Iteration = {it}")

# materialise only the basic features to disks
pd.DataFrame(audit_arr, columns=audit_columns).to_csv(audit_path, index=False)
Xf.to_csv(updated_basic_features_path, index=False)